import logging
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

import typer
from module_qc_data_tools import (
    get_layer_from_sn,
    get_sn_from_connectivity,
    outputDataFrame,
    save_dict_list,
)

from module_qc_tools.cli.globals import (
    CONTEXT_SETTINGS,
    OPTIONS,
    LogLevel,
)
from module_qc_tools.measurements.iv_measure import run
from module_qc_tools.utils.misc import (
    add_identifiers_metadata,
    copytree,
    load_config,
)
from module_qc_tools.utils.power_supply import power_supply

logger = logging.getLogger("measurement")

app = typer.Typer(context_settings=CONTEXT_SETTINGS)

# Taking the test-type from the script name which is the test-code in ProdDB.
TEST_TYPE = Path(__file__).stem


@app.command()
def main(
    config_path: Path = OPTIONS["config"],
    base_output_dir: Path = OPTIONS["output_dir"],
    module_connectivity: Optional[Path] = OPTIONS["module_connectivity"],
    _verbosity: LogLevel = OPTIONS["verbosity"],
    institution: str = OPTIONS["site"],
    use_calib_adc: bool = OPTIONS["use_calib_ADC"],
    save_local: bool = OPTIONS["save_local"],
    vdepl: float = OPTIONS["depl_volt"],
):
    timestart = datetime.now().strftime("%Y-%m-%d_%H%M%S")

    if use_calib_adc:
        logger.warning(
            "The --use-calib-adc was supplied but ADC calibration cannot be performed with this option. Will ignore and use multimeter."
        )

    # if -o option used, overwrite the default output directory
    output_dir = module_connectivity.parent if module_connectivity else base_output_dir

    if base_output_dir != Path("outputs"):
        output_dir = base_output_dir

    output_dir = output_dir.joinpath("Measurements", TEST_TYPE, timestart)
    # Make output directory and start log file
    output_dir.mkdir(parents=True, exist_ok=True)

    logger.info(f"Start {TEST_TYPE.replace('_',' ')}!")
    logger.info(f"TimeStart: {timestart}")

    logger.addHandler(logging.FileHandler(output_dir.joinpath("output.log")))

    config = load_config(config_path)

    if module_connectivity:
        config["yarr"]["connectivity"] = module_connectivity

    # Taking the module SN from YARR path to config in the connectivity file.
    module_serial = get_sn_from_connectivity(config["yarr"]["connectivity"])
    layer = get_layer_from_sn(module_serial)

    # create measurement config
    measurement_config = config["tasks"]["GENERAL"]
    measurement_config.update(config["tasks"][TEST_TYPE])

    # initialize hardware
    ps = power_supply(config["power_supply"])
    hv = power_supply(config["high_voltage"], name="high_voltage", is_hv=True)

    if hv.getPolarity() < 0:
        logger.info(
            f"HV polarity is {hv.getPolarity()} which is opposite to default (1)."
        )
        measurement_config["v_max"].update(
            (x, abs(y)) for x, y in measurement_config["v_max"].items()
        )

    try:
        data = run(config, ps, hv, layer)
    except KeyboardInterrupt:
        logger.info("KeyboardInterrupt")
    except Exception as err:
        logger.exception(err)
        raise typer.Exit(1) from err

    add_identifiers_metadata(data, module_serial, institution)
    data.add_meta_data("DepletionVoltage", vdepl)

    logger.info(
        "==================================Summary=================================="
    )
    alloutput = []

    logger.info("data: %s", data, extra={"markup": False})
    outputDF = outputDataFrame()
    outputDF._serialNumber = module_serial
    outputDF.set_test_type(TEST_TYPE)
    outputDF.set_results(data)
    alloutput += [outputDF.to_dict()]

    with TemporaryDirectory() as tmpdirname:
        save_dict_list(
            Path(tmpdirname).joinpath(f"{module_serial}.json"),
            alloutput,
        )

        # for now, set to false until localDB upload functionality implemented
        upload_failed = True
        upload_implemented = False

        if not save_local:
            # add in logic here to upload to localDB
            msg = "Not implemented yet"
            raise RuntimeError(msg)

        if upload_failed or save_local:
            copytree(tmpdirname, output_dir)
            if upload_failed and upload_implemented:
                logger.warning(
                    "The upload to localDB failed. Please fix and retry uploading the measurement output again."
                )

            logger.info(f"Writing output measurements in {output_dir}")

    logger.info("Done!")
    logger.info(f"TimeEnd: {datetime.now().strftime('%Y-%m-%d_%H%M%S')}")


if __name__ == "__main__":
    typer.run(main)
